REBOL [
	Title:		"SVG Demo"
	Owner:		"Ashley G. Trter"
	Version:	0.0.1
	Date:		21-Jun-2005
	Purpose:	"Loads and displays a resizeable SVG file."
	History: {
		0.0.1	Initial release
	}
	Notes: {
		Tested on very simple SVG icons
		Only a few basic styles / attributes / commands supported
		Does not handle sizes in units other than pixels (e.g. pt, in, cm, mm, etc)
		SVG path has an optional close command, "z" ... AGG shape equivalent auto-closes
		load-svg function needs to be totally refactored / optimized ... *sample only*
	}
]

;	The following commands are available for path data:
;
;		M = moveto
;		L = lineto
;		H = horizontal lineto
;		V = vertical lineto
;		C = curveto
;		S = smooth curveto
;		Q = quadratic Belzier curve
;		T = smooth quadratic Belzier curveto
;		A = elliptical Arc
;		Z = closepath

; conversion unit
;     *  "1pt" equals "1.25px" (and therefore 1.25 user units)
;     * "1pc" equals "15px" (and therefore 15 user units)
;     * "1mm" would be "3.543307px" (3.543307 user units)
;     * "1cm" equals "35.43307px" (and therefore 35.43307 user units)
;     * "1in" equals "90px" (and therefore 90 user units)


;print: none	; comment out this line to enable debug messages

group-blk: copy []
group-flag: none 

load-svg: function [svg-file [file! string!] size [pair!]] [
	id defs x y to-color to-byte draw-blk append-style svg-size scale-x scale-y
 	pair add-command token
][
	xml: either string? svg-file [parse-xml svg-file ][
		unless %.svg = suffix? svg-file [to error! "File has an invalid suffix!"]
		xml: parse-xml read svg-file
	]
	
	unless xml/3/1/1 = "svg" [to error! "Could not find SVG header!"]
	;unless find ["id" "xmlns"] xml/3/1/2/1 [to error! "Could not find ID header!"]
	;unless xml/3/1/3/1/1 = "defs" [to error! "Could not find DEFS header!"]
	
	id: xml/3/1/2
	defs: xml/3/1/3
	;probe defs
	;
	;	--- Parse SVG id
	;

	svg-size: either find ["32pt" "48pt" "72pt"] select id "width" [
		switch select id "width" [
			"72pt"	[120x120]
			"48pt"	[80x80]
			"32pt"	[60x60]
		]
	][
		as-pair to integer! any [select id "width" "100"] to integer! any [select id "height" "100"]
	]

	x: to integer! any [select id "x" "0"]
	y: to integer! any [select id "y" "0"]

	scale-x: size/x / svg-size/x
	scale-y: size/y / svg-size/y

	;scale-x: 1
	;scale-y: 1
	
	;
	;	--- Helper functions
	;

	to-color: func [s [string!]] [	; converts a string in the form "#FFFFFF" to a 4-byte tuple
		to tuple! load rejoin ["#{" next s "00}"]
	]

	to-byte: func [s [string!]] [	; converts a string with a value 0-1 to an inverted byte
		255 - to integer! 255 * to decimal! s
	]

	;
	;	--- Parse SVG defs
	;

	draw-blk: copy []

	append-style: function [
		command [string!] blk [block!]
	][
		x xy pen-color fill-color line-width mode size radius shape
		closed? matrix transf-command
	][
		xy: 0x0
		size: 0x0
		line-width: 1
		matrice: make block! []
		radius: none
		transf-command: none
		;gmatrix:[]
		
		foreach [attr val] blk [
			switch attr [
				"transform" [print "tranform have been found" 
					;probe val halt 
					val: parse val "(),"
					transf-command: first val
					;probe transf-command
					switch transf-command [
						"matrix" [ 
							foreach word val [
								if not find word "matrix"
								[ 
									insert tail matrice to-decimal word
								]
							]
						]
					]
				]
				"style" [
					foreach [attr val] parse val ":;" [
						switch/default attr [
						
							"font-size" [ ]
							"stroke" [
								switch/default first val [
									#"#" [pen-color: to-color val]
									#"n" [pen-color: none]
								][
									print ["Unknown stroke:" val]
								]
							]
							"stroke-width" [line-width: to decimal! val]
							"fill" [
								fill-color: switch/default first val [
									#"#" [to-color val]
									#"n" [none]
								][
									print ["Unknown fill value:" val]
									none
								]
							]
							"fill-rule" [
								mode: switch/default val [
									"evenodd"	['even-odd]
								][
									print ["Unknown fill-rule value:" val]
									none
								]
							]
							"stroke-opacity" [pen-color: any [pen-color 0.0.0.0] pen-color/4: to-byte val]
							"fill-opacity" [fill-color: any [fill-color 0.0.0.0] fill-color/4: to-byte val]
							"stroke-linejoin" [
								insert tail draw-blk switch/default val [
									"miter"		[compose [line-join miter]]
									"round"		[compose [line-join round]]
									"bevel"		[compose [line-join bevel]]
								][
									print ["Unknown stroke-linejoin value:" val]
									none
								]
							]
							"stroke-linecap" [
								insert tail draw-blk 'line-cap
								insert tail draw-blk to word! val
							]
						][
							print ["Unknown style:" attr]
						]
					]
				]
				"x"			[xy/x: scale-x * val]
				"y"			[xy/y: scale-y * val]
				"width"		[size/x: scale-x * val]
				"height"	[size/y: scale-y * val]
				"rx"		[print "rx"]
				"ry"		[radius: to decimal! val]
				
				"d"	[
					shape: copy []
					x: y: none
					closed?: false
					if find val "," [ ; adapattion routine to adapt inkscape format for d= code...
							print "INK SCAPE FORMAT FOUND"
							val: parse val ","
							b: ""
							forall val [
								b: either empty? b [val/1][rejoin [b " " val/1]]
							] 
							val: copy b b: none
	 					]
					pair: [set x number! set y number! (insert tail shape as-pair scale-x * x scale-y * y)]
					add-command: func [token [word!] command [word!]][
						insert tail shape either (first form token) = first uppercase form token
							[command][to-lit-word command]
					]
					parse load val [
						any [
							[set token 'M (add-command token 'move)   any pair] |
							[set token 'L (add-command token 'line)   any pair] |
							[set token 'H (add-command token 'hline)  any [
								set x number! (insert tail shape scale-x * x)]
							] |
							[set token 'V (add-command token 'vline)  any [
								set y number! (insert tail shape scale-y * y)]
							] |
							[set token 'C (add-command token 'curve)  any pair] |
							[set token 'S (add-command token 'curv)   any pair] |
							[set token 'Q (add-command token 'qcurve) any pair] |
							[set token 'T (add-command token 'qcurv)  any pair] |
							'z (closed?: true) |
							set token 1 skip (
								unless number? token [print ["Unknown path command:" token]]
							)
						]
					]
				]
; 				"d"	[
; 					shape: copy []
; 					x: none
; 					closed?: false
; 					if find val "," [ ; adapattion routine to adapt inkscape format for d= code...
; 						print "INK SCAPE FORMAT FOUND"
; 						val: parse val ","
; 						b: ""
; 						forall val [
; 							b: either empty? b [val/1][rejoin [b " " val/1]]
; 						] 
; 						val: copy b b: none
;  					]
; 					; SODIPODI SVG PLAIN TEXT FORMAT...
; 					foreach token load val [
; 						if all [x not number? token] [
; 								insert tail shape x * either token = 'V [scale-y][scale-x]
; 							x: none
; 						]
; 						
; 						switch/default token [
; 							M	[insert tail shape 'move]
; 							C	[insert tail shape 'curve]
; 							S   [insert tail shape 'curv]
; 							L	[insert tail shape 'line]
; 							Q   [insert tail shape 'qcurve]
; 							T   [insert tail shape 'qcurv]
; 							z	[closed?: true]
; 							A   [insert tail shape 'arc]
; 						][
; 							
; 							unless number? token [print ["Unknown path command:" token]]
; 							either x [insert tail shape as-pair x scale-y * token x: none] [x: scale-x * token]
; 							
; 						]
; 					]
; 				]
			]
		]
		insert tail draw-blk compose [
			pen (pen-color)
			fill-pen (fill-color)
			fill-rule (mode)
			line-width (line-width * min scale-x scale-y)
		]
		switch command [
			"rect" [
				insert tail draw-blk compose [box (xy) (xy + size)]
				if radius [insert tail draw-blk radius]
			]
			"path" [
				unless closed? [insert tail shape reduce ['move 0x0]]
				either transf-command <> none  [
					switch transf-command [
						"matrix" [
							either group-flag [
								insert tail group-blk  compose/deep/only [matrix (gmatrix) (to-word transf-command) (matrice) shape (shape) reset-matrix matrix (gmatrix)  ]
							][
								insert tail draw-blk compose/deep/only [ matrix (gmatrix) (to-word transf-command) (matrice) shape (shape)]
							]
							matrice: copy [] tranf-command: none
						]
					]
					
				][
						either group-flag [
							insert tail group-blk compose/only [shape (shape)]
						][
							insert tail draw-blk compose/only [shape (shape)]
						]
					]
				]
			"g" [ print "Write here how to handle G insertion to Draw block" 
					;insert tail gmatrix probe compose/only [(to-word transf-command) (matrice)]
					gmatrix:  copy matrice
					;probe gmatrix
				]
			]
	]
	linearGradient: make block! []
	
	append-gradient: function [cmd [string!] id  blk ]
	[ grad-settings ]
	[
		grad-settings: make block! []
		either id = "" [
			foreach [val attr] blk [
				probe val
				switch val [
					"id" [ id: copy attr ]
					"gradientTransform" [ insert tail grad-settings  compose/only [ grad-trans (attr)]]
					"gradientUnits" [insert tail grad-settings  compose/only [ grad-unit (attr)] ]
					"spreadMethod" [insert tail grad-settings  compose/only [ grad-spread (attr)]]
					"xlink:href" [ 
						insert tail grad-settings  compose/only [ grad-linked (next attr)]
					]
					"x1" [ insert tail grad-settings  compose/only [ x1 (attr)] ]
					"x2" [insert tail grad-settings  compose/only [ x2 (attr)]]
					"y1" [insert tail grad-settings  compose/only [ y1 (attr)]]
					"y2" [insert tail grad-settings  compose/only [ y2 (attr)]]
				]
			]
		][
			foreach dat blk [
				if block? dat [
					foreach [val attr] second dat [ 
						probe val
						switch val [
							"offset" [insert tail grad-settings compose/only [ grad-offs (attr)]]
							"style" [insert tail grad-settings compose/only [ grad-styl (attr)]]
							
						]
					]
				]
			]
			;halt
		]
		
		switch cmd [
			"linearGradient" [ 
				insert grad-settings compose/only [linear "" ]
				insert tail linearGradient compose/only [(id) (grad-settings)] ]
		]
	]
	
	linearGard: make block! []
	
	if defs/2/3 <> none [ 
		if defs/2/3/2/1 = "linearGradient" [
			Lingrad: copy defs/2/3
			foreach blk lingrad [
				if block? blk [
					either (first second blk) = "id" [
						probe blk
						append-gradient (first blk) (second second blk) (third  blk)
						; print "id-gradient" halt 
					][
						append-gradient (first blk) "" (second blk)
						;probe second blk
					]
				]
			]
			probe linearGradient
		]
	]
	;halt
	;probe defs/2/3/2/1]
	foreach blk defs [
		;probe second blk
		;print ""
		;print ""
		switch first blk [
			"rect"	[append-style first blk second blk]
			"path"	[append-style first blk second blk]
			"g"		[
					;	print "key word" probe first blk  
					;	print "matrix and style in G" probe second blk  
						group-flag: true
						append-style first blk second blk 
						;print "what to draw in G" probe third blk
						foreach blk2 third blk [
							;probe blk2
							switch first blk2 [ 
								"path" [append-style first blk2 second blk2]
								"rect" [append-style first blk2 second blk2]
							]
						]
						insert tail draw-blk compose/only [ push (group-blk)]
						group-flag: false
						group-blk: copy []
					]
		]
	]
	
	
    probe draw-blk
	draw-blk
]

view make face [
	offset:	100x100
	size:	200x200
	action:	request-file/filter/only "*.svg"
	text:	rejoin ["SVG Demo [" last split-path action "]"]
	data:	read action
	color:	white
	effect:	compose/only [draw (load-svg data size)]
	edge: font: para: none
	feel: make feel [
		redraw: func [face act pos] [
			if act = 'draw 
			[insert clear face/effect/draw load-svg face/data face/size]
		]
	]
	options: [resize]
]